@@ -1,23 +1,11 @@
 /*
- * Copyright 2015-2017 Milian Wolff <mail@milianw.de>
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Library General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public
- * License along with this program; if not, write to the
- * Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
- */
+    SPDX-FileCopyrightText: 2015-2020 Milian Wolff <mail@milianw.de>
+
+    SPDX-License-Identifier: LGPL-2.1-or-later
+*/
 
 #include "accumulatedtracedata.h"
+#include "analyze_config.h"
 
 #include <algorithm>
 #include <cassert>
@@ -26,12 +14,23 @@
 
 #include <boost/algorithm/string/predicate.hpp>
 #include <boost/iostreams/filter/gzip.hpp>
+#if ZSTD_FOUND
+#if BOOST_IOSTREAMS_HAS_ZSTD
+#include <boost/iostreams/filter/zstd.hpp>
+#else
+#include <boost-zstd/zstd.hpp>
+#endif
+#endif
 #include <boost/iostreams/filtering_stream.hpp>
 
+#include <boost/filesystem.hpp>
+
 #include "util/config.h"
 #include "util/linereader.h"
 #include "util/pointermap.h"
 
+#include "suppressions.h"
+
 #ifdef __GNUC__
 #define POTENTIALLY_UNUSED __attribute__((unused))
 #else
@@ -54,6 +53,32 @@ ostream& operator<<(ostream& out, const Index<Base> index)
     out << index.index;
     return out;
 }
+
+// boost's counter filter uses an int for the count which overflows for large streams; so replace it with a work alike.
+class byte_counter
+{
+public:
+    using char_type = char;
+    using category = boost::iostreams::multichar_input_filter_tag;
+
+    uint64_t bytes() const
+    {
+        return m_bytes;
+    }
+
+    template <typename Source>
+    std::streamsize read(Source& src, char* str, std::streamsize size)
+    {
+        auto const readsize = boost::iostreams::read(src, str, size);
+        if (readsize == -1)
+            return -1;
+        m_bytes += readsize;
+        return readsize;
+    }
+
+private:
+    uint64_t m_bytes = 0;
+};
 }
 
 AccumulatedTraceData::AccumulatedTraceData()
@@ -62,10 +87,13 @@ AccumulatedTraceData::AccumulatedTraceData()
     traces.reserve(65536);
     strings.reserve(4096);
     allocations.reserve(16384);
+    traceIndexToAllocationIndex.reserve(16384);
     stopIndices.reserve(4);
     opNewIpIndices.reserve(16);
 }
 
+AccumulatedTraceData::~AccumulatedTraceData() = default;
+
 const string& AccumulatedTraceData::stringify(const StringIndex stringId) const
 {
     if (!stringId || stringId.index > strings.size()) {
@@ -118,9 +146,16 @@ string AccumulatedTraceData::prettyFunction(const string& function) const
     return ret;
 }
 
-bool AccumulatedTraceData::read(const string& inputFile)
+bool AccumulatedTraceData::read(const string& inputFile, bool isReparsing)
 {
-    const bool isCompressed = boost::algorithm::ends_with(inputFile, ".gz");
+    return read(inputFile, FirstPass, isReparsing) && read(inputFile, SecondPass, isReparsing);
+}
+
+bool AccumulatedTraceData::read(const string& inputFile, const ParsePass pass, bool isReparsing)
+{
+    const bool isGzCompressed = boost::algorithm::ends_with(inputFile, ".gz");
+    const bool isZstdCompressed = boost::algorithm::ends_with(inputFile, ".zst");
+    const bool isCompressed = isGzCompressed || isZstdCompressed;
     ifstream file(inputFile, isCompressed ? ios_base::in | ios_base::binary : ios_base::in);
 
     if (!file.is_open()) {
@@ -129,37 +164,66 @@ bool AccumulatedTraceData::read(const string& inputFile)
     }
 
     boost::iostreams::filtering_istream in;
-    if (isCompressed) {
+    in.push(byte_counter()); // caution, ::read dependant on filter order
+    if (isGzCompressed) {
         in.push(boost::iostreams::gzip_decompressor());
+    } else if (isZstdCompressed) {
+#if ZSTD_FOUND
+        in.push(boost::iostreams::zstd_decompressor());
+#else
+        cerr << "Heaptrack was built without zstd support, cannot decompressed data file: " << inputFile << endl;
+        return false;
+#endif
     }
+    in.push(byte_counter()); // caution, ::read dependant on filter order
     in.push(file);
-    return read(in);
+
+    parsingState.fileSize = boost::filesystem::file_size(inputFile);
+
+    return read(in, pass, isReparsing);
 }
 
-bool AccumulatedTraceData::read(istream& in)
+bool AccumulatedTraceData::read(boost::iostreams::filtering_istream& in, const ParsePass pass, bool isReparsing)
 {
     LineReader reader;
     int64_t timeStamp = 0;
 
     vector<string> opNewStrings = {
         // 64 bit
-        "operator new(unsigned long)", "operator new[](unsigned long)",
+        "operator new(unsigned long)",
+        "operator new[](unsigned long)",
         // 32 bit
-        "operator new(unsigned int)", "operator new[](unsigned int)",
+        "operator new(unsigned int)",
+        "operator new[](unsigned int)",
     };
     vector<StringIndex> opNewStrIndices;
     opNewStrIndices.reserve(opNewStrings.size());
 
     vector<string> stopStrings = {"main", "__libc_start_main", "__static_initialization_and_destruction_0"};
 
-    const bool reparsing = totalTime != 0;
-    m_maxAllocationTraceIndex.index = 0;
+    const auto lastPeakCost = pass != FirstPass ? totalCost.peak : 0;
+    const auto lastPeakTime = pass != FirstPass ? peakTime : 0;
+
     totalCost = {};
     peakTime = 0;
-    systemInfo = {};
+    if (pass == FirstPass) {
+        if (!filterParameters.disableBuiltinSuppressions) {
+            suppressions = builtinSuppressions();
+        }
+
+        suppressions.resize(suppressions.size() + filterParameters.suppressions.size());
+        std::transform(filterParameters.suppressions.begin(), filterParameters.suppressions.end(), suppressions.begin(),
+                       [](const std::string& pattern) {
+                           return Suppression {pattern, 0, 0};
+                       });
+    }
     peakRSS = 0;
-    allocations.clear();
-    uint fileVersion = 0;
+    for (auto& allocation : allocations) {
+        allocation.clearCost();
+    }
+    unsigned int fileVersion = 0;
+    bool debuggeeEncountered = false;
+    bool inFilteredTime = !filterParameters.minTime;
 
     // required for backwards compatibility
     // newer versions handle this in heaptrack_interpret already
@@ -170,12 +234,31 @@ bool AccumulatedTraceData::read(istream& in)
     // allocations, i.e. when a deallocation follows with the same data
     uint64_t lastAllocationPtr = 0;
 
-    while (reader.getLine(in)) {
+    const auto uncompressedCount = in.component<byte_counter>(0);
+    const auto compressedCount = in.component<byte_counter>(in.size() - 2);
+
+    parsingState.pass = pass;
+    parsingState.reparsing = isReparsing;
+
+    while (timeStamp < filterParameters.maxTime && reader.getLine(in)) {
+        parsingState.readCompressedByte = compressedCount->bytes();
+        parsingState.readUncompressedByte = uncompressedCount->bytes();
+        parsingState.timestamp = timeStamp;
+
         if (reader.mode() == 's') {
-            if (reparsing) {
+            if (pass != FirstPass || isReparsing) {
                 continue;
             }
-            strings.push_back(reader.line().substr(2));
+            if (fileVersion >= 3) {
+                // read sized string directly
+                std::string string;
+                reader >> string;
+                strings.push_back(std::move(string));
+            } else {
+                // read remaining line as string, possibly including white spaces
+                strings.push_back(reader.line().substr(2));
+            }
+
             StringIndex index;
             index.index = strings.size();
 
@@ -191,7 +274,7 @@ bool AccumulatedTraceData::read(istream& in)
                 }
             }
         } else if (reader.mode() == 't') {
-            if (reparsing) {
+            if (pass != FirstPass || isReparsing) {
                 continue;
             }
             TraceNode node;
@@ -203,30 +286,40 @@ bool AccumulatedTraceData::read(istream& in)
             }
             traces.push_back(node);
         } else if (reader.mode() == 'i') {
-            if (reparsing) {
+            if (pass != FirstPass || isReparsing) {
                 continue;
             }
             InstructionPointer ip;
             reader >> ip.instructionPointer;
             reader >> ip.moduleIndex;
-            reader >> ip.functionIndex;
-            reader >> ip.fileIndex;
-            reader >> ip.line;
+            auto readFrame = [&reader](Frame* frame) {
+                return (reader >> frame->functionIndex) && (reader >> frame->fileIndex) && (reader >> frame->line);
+            };
+            if (readFrame(&ip.frame)) {
+                Frame inlinedFrame;
+                while (readFrame(&inlinedFrame)) {
+                    ip.inlined.push_back(inlinedFrame);
+                }
+            }
+
             instructionPointers.push_back(ip);
-            if (find(opNewStrIndices.begin(), opNewStrIndices.end(), ip.functionIndex) != opNewStrIndices.end()) {
+            if (find(opNewStrIndices.begin(), opNewStrIndices.end(), ip.frame.functionIndex) != opNewStrIndices.end()) {
                 IpIndex index;
                 index.index = instructionPointers.size();
                 opNewIpIndices.push_back(index);
             }
         } else if (reader.mode() == '+') {
+            if (!inFilteredTime) {
+                continue;
+            }
             AllocationInfo info;
-            AllocationIndex allocationIndex;
+            AllocationInfoIndex allocationIndex;
             if (fileVersion >= 1) {
-                if (!(reader >> allocationIndex.index)) {
-                    cerr << "failed to parse line: " << reader.line() << endl;
+                if (!(reader >> allocationIndex)) {
+                    cerr << "failed to parse line: " << reader.line() << ' ' << __LINE__ << endl;
                     continue;
                 } else if (allocationIndex.index >= allocationInfos.size()) {
-                    cerr << "allocation index out of bounds: " << allocationIndex.index
+                    cerr << "allocation index out of bounds: " << allocationIndex
                          << ", maximum is: " << allocationInfos.size() << endl;
                     continue;
                 }
@@ -234,39 +327,47 @@ bool AccumulatedTraceData::read(istream& in)
                 lastAllocationPtr = allocationIndex.index;
             } else { // backwards compatibility
                 uint64_t ptr = 0;
-                if (!(reader >> info.size) || !(reader >> info.traceIndex) || !(reader >> ptr)) {
-                    cerr << "failed to parse line: " << reader.line() << endl;
+                TraceIndex traceIndex;
+                if (!(reader >> info.size) || !(reader >> traceIndex) || !(reader >> ptr)) {
+                    cerr << "failed to parse line: " << reader.line() << ' ' << __LINE__ << endl;
                     continue;
                 }
-                if (allocationInfoSet.add(info.size, info.traceIndex, &allocationIndex)) {
+                info.allocationIndex = mapToAllocationIndex(traceIndex);
+                if (allocationInfoSet.add(info.size, traceIndex, &allocationIndex)) {
                     allocationInfos.push_back(info);
                 }
                 pointers.addPointer(ptr, allocationIndex);
                 lastAllocationPtr = ptr;
             }
 
-            auto& allocation = findAllocation(info.traceIndex);
-            allocation.leaked += info.size;
-            allocation.allocated += info.size;
-            ++allocation.allocations;
-            if (allocation.leaked > allocation.peak) {
-                allocation.peak = allocation.leaked;
+            if (pass != FirstPass) {
+                auto& allocation = allocations[info.allocationIndex.index];
+                allocation.leaked += info.size;
+                ++allocation.allocations;
+
+                handleAllocation(info, allocationIndex);
             }
 
             ++totalCost.allocations;
-            totalCost.allocated += info.size;
             totalCost.leaked += info.size;
             if (totalCost.leaked > totalCost.peak) {
                 totalCost.peak = totalCost.leaked;
                 peakTime = timeStamp;
-            }
 
-            handleAllocation(info, allocationIndex);
+                if (pass == SecondPass && totalCost.peak == lastPeakCost && peakTime == lastPeakTime) {
+                    for (auto& allocation : allocations) {
+                        allocation.peak = allocation.leaked;
+                    }
+                }
+            }
         } else if (reader.mode() == '-') {
-            AllocationIndex allocationInfoIndex;
+            if (!inFilteredTime) {
+                continue;
+            }
+            AllocationInfoIndex allocationInfoIndex;
             bool temporary = false;
             if (fileVersion >= 1) {
-                if (!(reader >> allocationInfoIndex.index)) {
+                if (!(reader >> allocationInfoIndex)) {
                     cerr << "failed to parse line: " << reader.line() << endl;
                     continue;
                 }
@@ -288,23 +389,31 @@ bool AccumulatedTraceData::read(istream& in)
             lastAllocationPtr = 0;
 
             const auto& info = allocationInfos[allocationInfoIndex.index];
-            auto& allocation = findAllocation(info.traceIndex);
-            allocation.leaked -= info.size;
             totalCost.leaked -= info.size;
             if (temporary) {
-                ++allocation.temporary;
                 ++totalCost.temporary;
             }
+
+            if (pass != FirstPass) {
+                auto& allocation = allocations[info.allocationIndex.index];
+                allocation.leaked -= info.size;
+                if (temporary) {
+                    ++allocation.temporary;
+                }
+            }
         } else if (reader.mode() == 'a') {
-            if (reparsing) {
+            if (pass != FirstPass || isReparsing) {
                 continue;
             }
             AllocationInfo info;
-            if (!(reader >> info.size) || !(reader >> info.traceIndex)) {
+            TraceIndex traceIndex;
+            if (!(reader >> info.size) || !(reader >> traceIndex)) {
                 cerr << "failed to parse line: " << reader.line() << endl;
                 continue;
             }
+            info.allocationIndex = mapToAllocationIndex(traceIndex);
             allocationInfos.push_back(info);
+
         } else if (reader.mode() == '#') {
             // comment or empty line
             continue;
@@ -314,21 +423,36 @@ bool AccumulatedTraceData::read(istream& in)
                 cerr << "Failed to read time stamp: " << reader.line() << endl;
                 continue;
             }
-            handleTimeStamp(timeStamp, newStamp);
+            inFilteredTime = newStamp >= filterParameters.minTime && newStamp <= filterParameters.maxTime;
+            if (inFilteredTime) {
+                handleTimeStamp(timeStamp, newStamp, false, pass);
+            }
             timeStamp = newStamp;
         } else if (reader.mode() == 'R') { // RSS timestamp
+            if (!inFilteredTime) {
+                continue;
+            }
             int64_t rss = 0;
             reader >> rss;
             if (rss > peakRSS) {
                 peakRSS = rss;
             }
         } else if (reader.mode() == 'X') {
-            handleDebuggee(reader.line().c_str() + 2);
+            if (debuggeeEncountered) {
+                cerr << "Duplicated debuggee entry - corrupt data file?" << endl;
+                return false;
+            }
+            debuggeeEncountered = true;
+            if (pass != FirstPass && !isReparsing) {
+                handleDebuggee(reader.line().c_str() + 2);
+            }
         } else if (reader.mode() == 'A') {
+            if (pass != FirstPass || isReparsing)
+                continue;
             totalCost = {};
             fromAttached = true;
         } else if (reader.mode() == 'v') {
-            uint heaptrackVersion = 0;
+            unsigned int heaptrackVersion = 0;
             reader >> heaptrackVersion;
             if (!(reader >> fileVersion) && heaptrackVersion == 0x010200) {
                 // backwards compatibility: before the 1.0.0, I actually
@@ -345,19 +469,31 @@ bool AccumulatedTraceData::read(istream& in)
                      << endl;
                 return false;
             }
+            if (fileVersion >= 3) {
+                reader.setExpectedSizedStrings(true);
+            }
         } else if (reader.mode() == 'I') { // system information
             reader >> systemInfo.pageSize;
             reader >> systemInfo.pages;
+        } else if (reader.mode() == 'S') { // embedded suppression
+            if (pass != FirstPass || filterParameters.disableEmbeddedSuppressions) {
+                continue;
+            }
+            auto suppression = parseSuppression(reader.line().substr(2));
+            if (!suppression.empty()) {
+                suppressions.push_back({std::move(suppression), 0, 0});
+            }
         } else {
             cerr << "failed to parse line: " << reader.line() << endl;
         }
     }
 
-    if (!reparsing) {
+    if (pass == FirstPass && !isReparsing) {
         totalTime = timeStamp + 1;
+        filterParameters.maxTime = totalTime;
     }
 
-    handleTimeStamp(timeStamp, totalTime);
+    handleTimeStamp(timeStamp, timeStamp + 1, true, pass);
 
     return true;
 }
@@ -378,7 +514,9 @@ vector<IndexT> sortedIndices(size_t numIndices, SortF sorter)
 
 vector<StringIndex> remapStrings(vector<string>& lhs, const vector<string>& rhs)
 {
-    unordered_map<string, StringIndex> stringRemapping;
+    tsl::robin_map<string, StringIndex> stringRemapping;
+
+    // insert known strings in lhs into the map for lookup below
     StringIndex stringIndex;
     {
         stringRemapping.reserve(lhs.size());
@@ -388,6 +526,9 @@ vector<StringIndex> remapStrings(vector<string>& lhs, const vector<string>& rhs)
         }
     }
 
+    // now insert the missing strings form rhs into lhs
+    // and create a remapped string vector, keeping the order
+    // of the strings in rhs, but mapping into the string vector from lhs
     vector<StringIndex> map;
     {
         map.reserve(rhs.size() + 1);
@@ -395,6 +536,8 @@ vector<StringIndex> remapStrings(vector<string>& lhs, const vector<string>& rhs)
         for (const auto& string : rhs) {
             auto it = stringRemapping.find(string);
             if (it == stringRemapping.end()) {
+                // a string that only occurs in rhs, but not lhs
+                // add it to lhs to make sure we can find it again later on
                 ++stringIndex.index;
                 lhs.push_back(string);
                 map.push_back(stringIndex);
@@ -406,11 +549,21 @@ vector<StringIndex> remapStrings(vector<string>& lhs, const vector<string>& rhs)
     return map;
 }
 
-template <typename T>
-inline const T& identity(const T& t)
+// replace by std::identity once we can leverage C++20
+struct identity
 {
-    return t;
-}
+    template <typename T>
+    const T& operator()(const T& t) const
+    {
+        return t;
+    }
+
+    template <typename T>
+    T operator()(T&& t) const
+    {
+        return std::move(t);
+    }
+};
 
 template <typename IpMapper>
 int compareTraceIndices(const TraceIndex& lhs, const AccumulatedTraceData& lhsData, const TraceIndex& rhs,
@@ -444,18 +597,44 @@ int compareTraceIndices(const TraceIndex& lhs, const AccumulatedTraceData& lhsDa
     return lhsIp.compareWithoutAddress(rhsIp) ? -1 : 1;
 }
 
+POTENTIALLY_UNUSED void printCost(const AllocationData& data)
+{
+    cerr << data.allocations << " (" << data.temporary << "), " << data.peak << " (" << data.leaked << ")\n";
+}
+
 POTENTIALLY_UNUSED void printTrace(const AccumulatedTraceData& data, TraceIndex index)
 {
     do {
         const auto trace = data.findTrace(index);
         const auto& ip = data.findIp(trace.ipIndex);
         cerr << index << " (" << trace.ipIndex << ", " << trace.parentIndex << ")" << '\t'
-             << data.stringify(ip.functionIndex) << " in " << data.stringify(ip.moduleIndex) << " at "
-             << data.stringify(ip.fileIndex) << ':' << ip.line << '\n';
+             << data.stringify(ip.frame.functionIndex) << " in " << data.stringify(ip.moduleIndex) << " at "
+             << data.stringify(ip.frame.fileIndex) << ':' << ip.frame.line << '\n';
+        for (const auto& inlined : ip.inlined) {
+            cerr << '\t' << data.stringify(inlined.functionIndex) << " at " << data.stringify(inlined.fileIndex) << ':'
+                 << inlined.line << '\n';
+        }
         index = trace.parentIndex;
     } while (index);
     cerr << "---\n";
 }
+
+template <class ForwardIt, class BinaryPredicateCompare, class BinaryOpReduce>
+ForwardIt inplace_unique_reduce(ForwardIt first, ForwardIt last, BinaryPredicateCompare cmp, BinaryOpReduce reduce)
+{
+    if (first == last)
+        return last;
+
+    ForwardIt result = first;
+    while (++first != last) {
+        if (cmp(*result, *first)) {
+            reduce(*result, *first);
+        } else if (++result != first) {
+            *result = std::move(*first);
+        }
+    }
+    return ++result;
+}
 }
 
 void AccumulatedTraceData::diff(const AccumulatedTraceData& base)
@@ -466,28 +645,22 @@ void AccumulatedTraceData::diff(const AccumulatedTraceData& base)
     systemInfo.pages -= base.systemInfo.pages;
     systemInfo.pageSize -= base.systemInfo.pageSize;
 
-    // step 1: sort trace indices of allocations for efficient lookup
-    // step 2: while at it, also merge equal allocations
-    vector<TraceIndex> allocationTraceNodes;
-    allocationTraceNodes.reserve(allocations.size());
-    for (auto it = allocations.begin(); it != allocations.end();) {
-        const auto& allocation = *it;
-        auto sortedIt =
-            lower_bound(allocationTraceNodes.begin(), allocationTraceNodes.end(), allocation.traceIndex,
-                        [this](const TraceIndex& lhs, const TraceIndex& rhs) -> bool {
-                            return compareTraceIndices(lhs, *this, rhs, *this, identity<InstructionPointer>) < 0;
-                        });
-        if (sortedIt == allocationTraceNodes.end()
-            || compareTraceIndices(allocation.traceIndex, *this, *sortedIt, *this, identity<InstructionPointer>) != 0) {
-            allocationTraceNodes.insert(sortedIt, allocation.traceIndex);
-            ++it;
-        } else if (*sortedIt != allocation.traceIndex) {
-            findAllocation(*sortedIt) += allocation;
-            it = allocations.erase(it);
-        } else {
-            ++it;
-        }
-    }
+    // step 1: sort allocations for efficient lookup and to prepare for merging equal allocations
+
+    std::sort(allocations.begin(), allocations.end(), [this](const Allocation& lhs, const Allocation& rhs) {
+        return compareTraceIndices(lhs.traceIndex, *this, rhs.traceIndex, *this, identity {}) < 0;
+    });
+
+    // step 2: now merge equal allocations
+
+    allocations.erase(inplace_unique_reduce(
+                          allocations.begin(), allocations.end(),
+                          [this](const Allocation& lhs, const Allocation& rhs) {
+                              return compareTraceIndices(lhs.traceIndex, *this, rhs.traceIndex, *this, identity {})
+                                  == 0;
+                          },
+                          [](Allocation& lhs, const Allocation& rhs) { lhs += rhs; }),
+                      allocations.end());
 
     // step 3: map string indices from rhs to lhs data
 
@@ -497,10 +670,17 @@ void AccumulatedTraceData::diff(const AccumulatedTraceData& base)
             index.index = stringMap[index.index].index;
         }
     };
-    auto remapIp = [&remapString](InstructionPointer ip) -> InstructionPointer {
+    auto remapFrame = [&remapString](Frame frame) -> Frame {
+        remapString(frame.functionIndex);
+        remapString(frame.fileIndex);
+        return frame;
+    };
+    auto remapIp = [&remapString, &remapFrame](InstructionPointer ip) -> InstructionPointer {
         remapString(ip.moduleIndex);
-        remapString(ip.functionIndex);
-        remapString(ip.fileIndex);
+        ip.frame = remapFrame(ip.frame);
+        for (auto& inlined : ip.inlined) {
+            inlined = remapFrame(inlined);
+        }
         return ip;
     };
 
@@ -561,33 +741,24 @@ void AccumulatedTraceData::diff(const AccumulatedTraceData& base)
     };
 
     // find an equivalent trace or copy the data over if it does not exist yet
-    // a trace is equivalent if the complete backtrace has equal
-    // InstructionPointer
+    // a trace is equivalent if the complete backtrace has equal InstructionPointer
     // data while ignoring the actual pointer address
-    auto remapTrace = [&base, &allocationTraceNodes, this, remapIp, copyTrace](TraceIndex rhsIndex) -> TraceIndex {
-        if (!rhsIndex) {
-            return rhsIndex;
-        }
+    for (const auto& rhsAllocation : base.allocations) {
 
-        auto it = lower_bound(allocationTraceNodes.begin(), allocationTraceNodes.end(), rhsIndex,
-                              [&base, this, remapIp](const TraceIndex& lhs, const TraceIndex& rhs) -> bool {
-                                  return compareTraceIndices(lhs, *this, rhs, base, remapIp) < 0;
+        assert(rhsAllocation.traceIndex);
+        auto it = lower_bound(allocations.begin(), allocations.end(), rhsAllocation.traceIndex,
+                              [&base, this, remapIp](const Allocation& lhs, const TraceIndex& rhs) -> bool {
+                                  return compareTraceIndices(lhs.traceIndex, *this, rhs, base, remapIp) < 0;
                               });
 
-        if (it != allocationTraceNodes.end() && compareTraceIndices(*it, *this, rhsIndex, base, remapIp) == 0) {
-            return *it;
+        if (it == allocations.end()
+            || compareTraceIndices(it->traceIndex, *this, rhsAllocation.traceIndex, base, remapIp) != 0) {
+            Allocation lhsAllocation;
+            lhsAllocation.traceIndex = copyTrace(rhsAllocation.traceIndex);
+            it = allocations.insert(it, lhsAllocation);
         }
 
-        TraceIndex ret = copyTrace(rhsIndex);
-        allocationTraceNodes.insert(it, ret);
-        return ret;
-    };
-
-    for (const auto& rhsAllocation : base.allocations) {
-        const auto lhsTrace = remapTrace(rhsAllocation.traceIndex);
-        assert(base.findIp(base.findTrace(rhsAllocation.traceIndex).ipIndex)
-                   .equalWithoutAddress(findIp(findTrace(lhsTrace).ipIndex)));
-        findAllocation(lhsTrace) -= rhsAllocation;
+        (*it) -= rhsAllocation;
     }
 
     // step 5: remove allocations that don't show any differences
@@ -595,41 +766,49 @@ void AccumulatedTraceData::diff(const AccumulatedTraceData& base)
     //         we can still end up with merged backtraces that have a total
     //         of 0, but different "tails" of different origin with non-zero cost
     allocations.erase(remove_if(allocations.begin(), allocations.end(),
-                                [](const Allocation& allocation) -> bool { return allocation == AllocationData(); }),
+                                [&](const Allocation& allocation) -> bool { return allocation == AllocationData(); }),
                       allocations.end());
 }
 
-Allocation& AccumulatedTraceData::findAllocation(const TraceIndex traceIndex)
+AllocationIndex AccumulatedTraceData::mapToAllocationIndex(const TraceIndex traceIndex)
 {
+    AllocationIndex allocationIndex;
     if (traceIndex < m_maxAllocationTraceIndex) {
         // only need to search when the trace index is previously known
-        auto it = lower_bound(allocations.begin(), allocations.end(), traceIndex,
-                              [](const Allocation& allocation, const TraceIndex traceIndex) -> bool {
-                                  return allocation.traceIndex < traceIndex;
-                              });
-        if (it == allocations.end() || it->traceIndex != traceIndex) {
-            Allocation allocation;
-            allocation.traceIndex = traceIndex;
-            it = allocations.insert(it, allocation);
+        auto it = lower_bound(traceIndexToAllocationIndex.begin(), traceIndexToAllocationIndex.end(), traceIndex,
+                              [](const pair<TraceIndex, AllocationIndex>& indexMap,
+                                 const TraceIndex traceIndex) -> bool { return indexMap.first < traceIndex; });
+        if (it != traceIndexToAllocationIndex.end() && it->first == traceIndex) {
+            return it->second;
         }
-        return *it;
-    } else if (traceIndex == m_maxAllocationTraceIndex && !allocations.empty()) {
-        // reuse the last allocation
-        assert(allocations.back().traceIndex == traceIndex);
-    } else {
-        // actually a new allocation
+        // new allocation
+        allocationIndex.index = allocations.size();
+        traceIndexToAllocationIndex.insert(it, make_pair(traceIndex, allocationIndex));
         Allocation allocation;
         allocation.traceIndex = traceIndex;
         allocations.push_back(allocation);
+    } else if (traceIndex == m_maxAllocationTraceIndex && !allocations.empty()) {
+        // reuse the last allocation
+        assert(allocations[m_maxAllocationIndex.index].traceIndex == traceIndex);
+        allocationIndex = m_maxAllocationIndex;
+    } else {
+        // new allocation
+        allocationIndex.index = allocations.size();
+        traceIndexToAllocationIndex.push_back(make_pair(traceIndex, allocationIndex));
+        Allocation allocation;
+        allocation.traceIndex = traceIndex;
+        m_maxAllocationIndex.index = allocations.size();
+        allocations.push_back(allocation);
         m_maxAllocationTraceIndex = traceIndex;
     }
-    return allocations.back();
+    return allocationIndex;
 }
 
-InstructionPointer AccumulatedTraceData::findIp(const IpIndex ipIndex) const
+const InstructionPointer& AccumulatedTraceData::findIp(const IpIndex ipIndex) const
 {
+    static const InstructionPointer invalid;
     if (!ipIndex || ipIndex.index > instructionPointers.size()) {
-        return {};
+        return invalid;
     } else {
         return instructionPointers[ipIndex.index - 1];
     }
@@ -648,3 +827,122 @@ bool AccumulatedTraceData::isStopIndex(const StringIndex index) const
 {
     return find(stopIndices.begin(), stopIndices.end(), index) != stopIndices.end();
 }
+
+struct SuppressionStringMatch
+{
+    static constexpr auto NO_MATCH = std::numeric_limits<std::size_t>::max();
+
+    SuppressionStringMatch(std::size_t index = NO_MATCH)
+        : suppressionIndex(index)
+    {
+    }
+
+    explicit operator bool() const
+    {
+        return suppressionIndex != NO_MATCH;
+    }
+
+    std::size_t suppressionIndex;
+};
+
+void AccumulatedTraceData::applyLeakSuppressions()
+{
+    totalLeakedSuppressed = 0;
+
+    if (suppressions.empty()) {
+        return;
+    }
+
+    // match all strings once against all suppression rules
+    bool hasAnyMatch = false;
+    std::vector<SuppressionStringMatch> suppressedStrings(strings.size());
+    std::transform(strings.begin(), strings.end(), suppressedStrings.begin(), [&](const auto& string) {
+        auto it = std::find_if(suppressions.begin(), suppressions.end(), [&string](const Suppression& suppression) {
+            return matchesSuppression(suppression.pattern, string);
+        });
+        if (it == suppressions.end()) {
+            return SuppressionStringMatch();
+        } else {
+            hasAnyMatch = true;
+            return SuppressionStringMatch(static_cast<std::size_t>(std::distance(suppressions.begin(), it)));
+        }
+    });
+    if (!hasAnyMatch) {
+        // nothing matched the suppressions, we can return early
+        return;
+    }
+
+    auto isSuppressedString = [&suppressedStrings](StringIndex index) {
+        if (index && index.index <= suppressedStrings.size()) {
+            return suppressedStrings[index.index - 1];
+        } else {
+            return SuppressionStringMatch();
+        }
+    };
+    auto isSuppressedFrame = [&isSuppressedString](Frame frame) {
+        auto match = isSuppressedString(frame.functionIndex);
+        if (match) {
+            return match;
+        }
+        return isSuppressedString(frame.fileIndex);
+    };
+
+    // now match all instruction pointers against the suppressed strings
+    std::vector<SuppressionStringMatch> suppressedIps(instructionPointers.size());
+    std::transform(instructionPointers.begin(), instructionPointers.end(), suppressedIps.begin(), [&](const auto& ip) {
+        auto match = isSuppressedString(ip.moduleIndex);
+        if (match) {
+            return match;
+        }
+        match = isSuppressedFrame(ip.frame);
+        if (match) {
+            return match;
+        }
+        for (const auto& inlined : ip.inlined) {
+            match = isSuppressedFrame(inlined);
+            if (match) {
+                return match;
+            }
+        }
+        return SuppressionStringMatch();
+    });
+    suppressedStrings = {};
+    auto isSuppressedIp = [&suppressedIps](IpIndex index) {
+        if (index && index.index <= suppressedIps.size()) {
+            return suppressedIps[index.index - 1];
+        }
+        return SuppressionStringMatch();
+    };
+
+    // now match all trace indices against the suppressed instruction pointers
+    std::vector<SuppressionStringMatch> suppressedTraces(traces.size());
+    auto isSuppressedTrace = [&suppressedTraces](TraceIndex index) {
+        if (index && index.index <= suppressedTraces.size()) {
+            return suppressedTraces[index.index - 1];
+        }
+        return SuppressionStringMatch();
+    };
+    std::transform(traces.begin(), traces.end(), suppressedTraces.begin(), [&](const auto& trace) {
+        auto match = isSuppressedTrace(trace.parentIndex);
+        if (match) {
+            return match;
+        }
+        return isSuppressedIp(trace.ipIndex);
+    });
+    suppressedIps = {};
+
+    // now finally zero all the matching allocations
+    for (auto& allocation : allocations) {
+        auto match = isSuppressedTrace(allocation.traceIndex);
+        if (match) {
+            totalLeakedSuppressed += allocation.leaked;
+
+            auto& suppression = suppressions[match.suppressionIndex];
+            ++suppression.matches;
+            suppression.leaked += allocation.leaked;
+
+            totalCost.leaked -= allocation.leaked;
+            allocation.leaked = 0;
+        }
+    }
+}
