@@ -1,28 +1,17 @@
 /*
- * Copyright 2015-2017 Milian Wolff <mail@milianw.de>
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Library General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public
- * License along with this program; if not, write to the
- * Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
- */
+    SPDX-FileCopyrightText: 2015-2019 Milian Wolff <mail@milianw.de>
+
+    SPDX-License-Identifier: LGPL-2.1-or-later
+*/
 
 #include "flamegraph.h"
 
 #include <cmath>
 
 #include <QAction>
+#include <QApplication>
 #include <QCheckBox>
+#include <QClipboard>
 #include <QComboBox>
 #include <QCursor>
 #include <QDebug>
@@ -32,6 +21,9 @@
 #include <QGraphicsScene>
 #include <QGraphicsView>
 #include <QLabel>
+#include <QLineEdit>
+#include <QMenu>
+#include <QPushButton>
 #include <QStyleOption>
 #include <QToolTip>
 #include <QVBoxLayout>
@@ -42,49 +34,68 @@
 #include <KStandardAction>
 #include <ThreadWeaver/ThreadWeaver>
 
+#include "resultdata.h"
+#include "util.h"
+
 enum CostType
 {
     Allocations,
     Temporary,
     Peak,
     Leaked,
-    Allocated
 };
 Q_DECLARE_METATYPE(CostType)
 
+namespace {
+enum SearchMatchType
+{
+    NoSearch,
+    NoMatch,
+    DirectMatch,
+    ChildMatch
+};
+}
+
 class FrameGraphicsItem : public QGraphicsRectItem
 {
 public:
-    FrameGraphicsItem(const qint64 cost, CostType costType, const QString& function,
-                      FrameGraphicsItem* parent = nullptr);
-    FrameGraphicsItem(const qint64 cost, const QString& function, FrameGraphicsItem* parent);
+    FrameGraphicsItem(const qint64 cost, CostType costType, const Symbol& symbol,
+                      std::shared_ptr<const ResultData> resultData, FrameGraphicsItem* parent = nullptr);
+    FrameGraphicsItem(const qint64 cost, const Symbol& symbol, std::shared_ptr<const ResultData> resultData,
+                      FrameGraphicsItem* parent);
 
     qint64 cost() const;
     void setCost(qint64 cost);
-    QString function() const;
+    Symbol symbol() const;
 
     void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget = nullptr) override;
 
     QString description() const;
 
+    bool match(const QString& searchValue) const;
+    void setSearchMatchType(SearchMatchType matchType);
+
 protected:
     void hoverEnterEvent(QGraphicsSceneHoverEvent* event) override;
     void hoverLeaveEvent(QGraphicsSceneHoverEvent* event) override;
 
 private:
+    std::shared_ptr<const ResultData> m_resultData;
     qint64 m_cost;
-    QString m_function;
+    Symbol m_symbol;
     CostType m_costType;
     bool m_isHovered;
+    SearchMatchType m_searchMatch = NoSearch;
 };
 
 Q_DECLARE_METATYPE(FrameGraphicsItem*)
 
-FrameGraphicsItem::FrameGraphicsItem(const qint64 cost, CostType costType, const QString& function,
-                                     FrameGraphicsItem* parent)
+FrameGraphicsItem::FrameGraphicsItem(const qint64 cost, CostType costType, const Symbol& symbol,
+                                     std::shared_ptr<const ResultData> resultData, FrameGraphicsItem* parent)
     : QGraphicsRectItem(parent)
+    , m_resultData(std::move(resultData))
     , m_cost(cost)
-    , m_function(function)
+    , m_symbol(symbol)
     , m_costType(costType)
     , m_isHovered(false)
 {
@@ -92,8 +103,9 @@ FrameGraphicsItem::FrameGraphicsItem(const qint64 cost, CostType costType, const
     setAcceptHoverEvents(true);
 }
 
-FrameGraphicsItem::FrameGraphicsItem(const qint64 cost, const QString& function, FrameGraphicsItem* parent)
-    : FrameGraphicsItem(cost, parent->m_costType, function, parent)
+FrameGraphicsItem::FrameGraphicsItem(const qint64 cost, const Symbol& symbol,
+                                     std::shared_ptr<const ResultData> resultData, FrameGraphicsItem* parent)
+    : FrameGraphicsItem(cost, parent->m_costType, symbol, std::move(resultData), parent)
 {
 }
 
@@ -107,30 +119,36 @@ void FrameGraphicsItem::setCost(qint64 cost)
     m_cost = cost;
 }
 
-QString FrameGraphicsItem::function() const
+Symbol FrameGraphicsItem::symbol() const
 {
-    return m_function;
+    return m_symbol;
 }
 
 void FrameGraphicsItem::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* /*widget*/)
 {
-    if (isSelected() || m_isHovered) {
+    if (isSelected() || m_isHovered || m_searchMatch == DirectMatch) {
         auto selectedColor = brush().color();
         selectedColor.setAlpha(255);
         painter->fillRect(rect(), selectedColor);
-    } else {
+    } else if (m_searchMatch == NoMatch) {
+        auto noMatchColor = brush().color();
+        noMatchColor.setAlpha(50);
+        painter->fillRect(rect(), noMatchColor);
+    } else { // default, when no search is running, or a sub-item is matched
         painter->fillRect(rect(), brush());
     }
 
     const QPen oldPen = painter->pen();
     auto pen = oldPen;
-    pen.setColor(brush().color());
-    if (isSelected()) {
-        pen.setWidth(2);
+    if (m_searchMatch != NoMatch) {
+        pen.setColor(brush().color());
+        if (isSelected()) {
+            pen.setWidth(2);
+        }
+        painter->setPen(pen);
+        painter->drawRect(rect());
+        painter->setPen(oldPen);
     }
-    painter->setPen(pen);
-    painter->drawRect(rect());
-    painter->setPen(oldPen);
 
     const int margin = 4;
     const int width = rect().width() - 2 * margin;
@@ -139,25 +157,66 @@ void FrameGraphicsItem::paint(QPainter* painter, const QStyleOptionGraphicsItem*
         return;
     }
 
-    const int height = rect().height();
+    if (m_searchMatch == NoMatch) {
+        auto color = oldPen.color();
+        color.setAlpha(125);
+        pen.setColor(color);
+        painter->setPen(pen);
+    }
 
+    auto label = [this]() {
+        if (m_symbol.isValid()) {
+            return m_resultData->string(m_symbol.functionId);
+        }
+
+        // root
+        switch (m_costType) {
+        case Allocations:
+            return i18n("%1 allocations in total", m_cost);
+        case Temporary:
+            return i18n("%1 temporary allocations in total", m_cost);
+        case Peak:
+            return i18n("%1 peak memory consumption", Util::formatBytes(m_cost));
+        case Leaked:
+            return i18n("%1 leaked in total", Util::formatBytes(m_cost));
+        }
+        Q_UNREACHABLE();
+    }();
+
+    const int height = rect().height();
     painter->drawText(margin + rect().x(), rect().y(), width, height,
                       Qt::AlignVCenter | Qt::AlignLeft | Qt::TextSingleLine,
-                      option->fontMetrics.elidedText(m_function, Qt::ElideRight, width));
+                      option->fontMetrics.elidedText(label, Qt::ElideRight, width));
+
+    if (m_searchMatch == NoMatch) {
+        painter->setPen(oldPen);
+    }
 }
 
 void FrameGraphicsItem::hoverEnterEvent(QGraphicsSceneHoverEvent* event)
 {
     QGraphicsRectItem::hoverEnterEvent(event);
     m_isHovered = true;
+    update();
+}
+
+void FrameGraphicsItem::hoverLeaveEvent(QGraphicsSceneHoverEvent* event)
+{
+    QGraphicsRectItem::hoverLeaveEvent(event);
+    m_isHovered = false;
+    update();
 }
 
 QString FrameGraphicsItem::description() const
 {
+    const auto symbol = Util::toString(m_symbol, *m_resultData, Util::Short);
+
     // we build the tooltip text on demand, which is much faster than doing that
     // for potentially thousands of items when we load the data
-    QString tooltip;
-    KFormat format;
+    if (!parentItem()) {
+        return symbol;
+    }
+
     qint64 totalCost = 0;
     {
         auto item = this;
@@ -166,46 +225,48 @@ QString FrameGraphicsItem::description() const
         }
         totalCost = item->cost();
     }
-    const auto fraction = QString::number(double(m_cost) * 100. / totalCost, 'g', 3);
-    const auto function = QString(QLatin1String("<span style='font-family:monospace'>") + m_function.toHtmlEscaped()
-                                  + QLatin1String("</span>"));
-    if (!parentItem()) {
-        return function;
-    }
+    const auto fraction = Util::formatCostRelative(m_cost, totalCost);
 
+    QString tooltip;
     switch (m_costType) {
     case Allocations:
         tooltip = i18nc("%1: number of allocations, %2: relative number, %3: function label",
-                        "%1 (%2%) allocations in %3 and below.", m_cost, fraction, function);
+                        "%1 (%2%) allocations in %3 and below.", m_cost, fraction, symbol);
         break;
     case Temporary:
         tooltip = i18nc("%1: number of temporary allocations, %2: relative number, "
                         "%3 function label",
-                        "%1 (%2%) temporary allocations in %3 and below.", m_cost, fraction, function);
+                        "%1 (%2%) temporary allocations in %3 and below.", m_cost, fraction, symbol);
         break;
     case Peak:
-        tooltip =
-            i18nc("%1: peak consumption in bytes, %2: relative number, %3: "
-                  "function label",
-                  "%1 (%2%) peak consumption in %3 and below.", format.formatByteSize(m_cost), fraction, function);
+        tooltip = i18nc("%1: peak consumption in bytes, %2: relative number, %3: "
+                        "function label",
+                        "%1 (%2%) contribution to peak consumption in %3 and below.", Util::formatBytes(m_cost),
+                        fraction, symbol);
         break;
     case Leaked:
         tooltip = i18nc("%1: leaked bytes, %2: relative number, %3: function label", "%1 (%2%) leaked in %3 and below.",
-                        format.formatByteSize(m_cost), fraction, function);
-        break;
-    case Allocated:
-        tooltip = i18nc("%1: allocated bytes, %2: relative number, %3: function label",
-                        "%1 (%2%) allocated in %3 and below.", format.formatByteSize(m_cost), fraction, function);
+                        Util::formatBytes(m_cost), fraction, symbol);
         break;
     }
 
     return tooltip;
 }
 
-void FrameGraphicsItem::hoverLeaveEvent(QGraphicsSceneHoverEvent* event)
+bool FrameGraphicsItem::match(const QString& searchValue) const
 {
-    QGraphicsRectItem::hoverLeaveEvent(event);
-    m_isHovered = false;
+    auto match = [&](StringIndex index) {
+        return m_resultData->string(index).contains(searchValue, Qt::CaseInsensitive);
+    };
+    return match(m_symbol.functionId) || match(m_symbol.moduleId);
+}
+
+void FrameGraphicsItem::setSearchMatchType(SearchMatchType matchType)
+{
+    if (m_searchMatch != matchType) {
+        m_searchMatch = matchType;
+        update();
+    }
 }
 
 namespace {
@@ -253,11 +314,11 @@ void layoutItems(FrameGraphicsItem* parent)
     }
 }
 
-FrameGraphicsItem* findItemByFunction(const QList<QGraphicsItem*>& items, const QString& function)
+FrameGraphicsItem* findItemBySymbol(const QList<QGraphicsItem*>& items, const Symbol& symbol)
 {
     foreach (auto item_, items) {
         auto item = static_cast<FrameGraphicsItem*>(item_);
-        if (item->function() == function) {
+        if (item->symbol() == symbol) {
             return item;
         }
     }
@@ -267,25 +328,25 @@ FrameGraphicsItem* findItemByFunction(const QList<QGraphicsItem*>& items, const
 /**
  * Convert the top-down graph into a tree of FrameGraphicsItem.
  */
-void toGraphicsItems(const QVector<RowData>& data, FrameGraphicsItem* parent, int64_t AllocationData::*member,
-                     const double costThreshold, bool collapseRecursion)
+void toGraphicsItems(const std::shared_ptr<const ResultData>& resultData, const QVector<RowData>& data,
+                     FrameGraphicsItem* parent, int64_t AllocationData::*member, const double costThreshold,
+                     bool collapseRecursion)
 {
     foreach (const auto& row, data) {
-        if (collapseRecursion && row.location->function != unresolvedFunctionName()
-            && row.location->function == parent->function())
-        {
+        if (collapseRecursion && row.symbol.functionId && row.symbol == parent->symbol()) {
+            toGraphicsItems(resultData, row.children, parent, member, costThreshold, collapseRecursion);
             continue;
         }
-        auto item = findItemByFunction(parent->childItems(), row.location->function);
+        auto item = findItemBySymbol(parent->childItems(), row.symbol);
         if (!item) {
-            item = new FrameGraphicsItem(row.cost.*member, row.location->function, parent);
+            item = new FrameGraphicsItem(row.cost.*member, row.symbol, resultData, parent);
             item->setPen(parent->pen());
             item->setBrush(brush());
         } else {
             item->setCost(item->cost() + row.cost.*member);
         }
         if (item->cost() > costThreshold) {
-            toGraphicsItems(row.children, item, member, costThreshold, collapseRecursion);
+            toGraphicsItems(resultData, row.children, item, member, costThreshold, collapseRecursion);
         }
     }
 }
@@ -301,83 +362,89 @@ int64_t AllocationData::*memberForType(CostType type)
         return &AllocationData::peak;
     case Leaked:
         return &AllocationData::leaked;
-    case Allocated:
-        return &AllocationData::allocated;
     }
     Q_UNREACHABLE();
 }
 
-FrameGraphicsItem* parseData(const QVector<RowData>& topDownData, CostType type, double costThreshold,
-                             bool collapseRecursion)
+FrameGraphicsItem* parseData(const TreeData& data, CostType type, double costThreshold, bool collapseRecursion)
 {
     auto member = memberForType(type);
 
-    double totalCost = 0;
-    foreach (const auto& frame, topDownData) {
-        totalCost += frame.cost.*member;
-    }
+    const auto totalCost = data.resultData->totalCosts().*member;
 
     KColorScheme scheme(QPalette::Active);
     const QPen pen(scheme.foreground().color());
 
-    KFormat format;
-    QString label;
-    switch (type) {
-    case Allocations:
-        label = i18n("%1 allocations in total", totalCost);
-        break;
-    case Temporary:
-        label = i18n("%1 temporary allocations in total", totalCost);
-        break;
-    case Peak:
-        label = i18n("%1 peak consumption in total", format.formatByteSize(totalCost));
-        break;
-    case Leaked:
-        label = i18n("%1 leaked in total", format.formatByteSize(totalCost));
-        break;
-    case Allocated:
-        label = i18n("%1 allocated in total", format.formatByteSize(totalCost));
-        break;
-    }
-    auto rootItem = new FrameGraphicsItem(totalCost, type, label);
+    auto rootItem = new FrameGraphicsItem(totalCost, type, {}, data.resultData);
     rootItem->setBrush(scheme.background());
     rootItem->setPen(pen);
-    toGraphicsItems(topDownData, rootItem, member, totalCost * costThreshold / 100., collapseRecursion);
+    toGraphicsItems(data.resultData, data.rows, rootItem, member, totalCost * costThreshold / 100., collapseRecursion);
     return rootItem;
 }
+
+struct SearchResults
+{
+    SearchMatchType matchType = NoMatch;
+    qint64 directCost = 0;
+};
+
+SearchResults applySearch(FrameGraphicsItem* item, const QString& searchValue)
+{
+    SearchResults result;
+    if (searchValue.isEmpty()) {
+        result.matchType = NoSearch;
+    } else if (item->match(searchValue)) {
+        result.directCost += item->cost();
+        result.matchType = DirectMatch;
+    }
+
+    // recurse into the child items, we always need to update all items
+    for (auto* child : item->childItems()) {
+        auto* childFrame = static_cast<FrameGraphicsItem*>(child);
+        auto childMatch = applySearch(childFrame, searchValue);
+        if (result.matchType != DirectMatch
+            && (childMatch.matchType == DirectMatch || childMatch.matchType == ChildMatch)) {
+            result.matchType = ChildMatch;
+            result.directCost += childMatch.directCost;
+        }
+    }
+
+    item->setSearchMatchType(result.matchType);
+    return result;
+}
 }
 
-FlameGraph::FlameGraph(QWidget* parent, Qt::WindowFlags flags)
-    : QWidget(parent, flags)
+FlameGraph::FlameGraph(QWidget* parent)
+    : QWidget(parent)
     , m_costSource(new QComboBox(this))
     , m_scene(new QGraphicsScene(this))
     , m_view(new QGraphicsView(this))
     , m_displayLabel(new QLabel)
+    , m_searchResultsLabel(new QLabel)
 {
     qRegisterMetaType<FrameGraphicsItem*>();
 
-    m_costSource->addItem(i18n("Allocations"), QVariant::fromValue(Allocations));
-    m_costSource->setItemData(0, i18n("Show a flame graph over the number of allocations triggered by "
-                                      "functions in your code."),
-                              Qt::ToolTipRole);
-    m_costSource->addItem(i18n("Temporary Allocations"), QVariant::fromValue(Temporary));
-    m_costSource->setItemData(1, i18n("Show a flame graph over the number of temporary allocations "
-                                      "triggered by functions in your code. "
-                                      "Allocations are marked as temporary when they are immediately "
-                                      "followed by their deallocation."),
-                              Qt::ToolTipRole);
-    m_costSource->addItem(i18n("Peak Consumption"), QVariant::fromValue(Peak));
-    m_costSource->setItemData(2, i18n("Show a flame graph over the peak heap "
-                                      "memory consumption of your application."),
+    m_costSource->addItem(i18n("Memory Peak"), QVariant::fromValue(Peak));
+    m_costSource->setItemData(2,
+                              i18n("Show a flame graph over the contributions to the peak heap "
+                                   "memory consumption of your application."),
                               Qt::ToolTipRole);
     m_costSource->addItem(i18n("Leaked"), QVariant::fromValue(Leaked));
-    m_costSource->setItemData(3, i18n("Show a flame graph over the leaked heap memory of your application. "
-                                      "Memory is considered to be leaked when it never got deallocated. "),
+    m_costSource->setItemData(3,
+                              i18n("Show a flame graph over the leaked heap memory of your application. "
+                                   "Memory is considered to be leaked when it never got deallocated. "),
                               Qt::ToolTipRole);
-    m_costSource->addItem(i18n("Allocated"), QVariant::fromValue(Allocated));
-    m_costSource->setItemData(4, i18n("Show a flame graph over the total memory allocated by functions in "
-                                      "your code. "
-                                      "This aggregates all memory allocations and ignores deallocations."),
+    m_costSource->addItem(i18n("Allocations"), QVariant::fromValue(Allocations));
+    m_costSource->setItemData(0,
+                              i18n("Show a flame graph over the number of allocations triggered by "
+                                   "functions in your code."),
+                              Qt::ToolTipRole);
+    m_costSource->addItem(i18n("Temporary Allocations"), QVariant::fromValue(Temporary));
+    m_costSource->setItemData(1,
+                              i18n("Show a flame graph over the number of temporary allocations "
+                                   "triggered by functions in your code. "
+                                   "Allocations are marked as temporary when they are immediately "
+                                   "followed by their deallocation."),
                               Qt::ToolTipRole);
     connect(m_costSource, static_cast<void (QComboBox::*)(int)>(&QComboBox::currentIndexChanged), this,
             &FlameGraph::showData);
@@ -389,9 +456,17 @@ FlameGraph::FlameGraph(QWidget* parent, Qt::WindowFlags flags)
     m_view->viewport()->setMouseTracking(true);
     m_view->setFont(QFont(QStringLiteral("monospace")));
 
-    auto bottomUpCheckbox = new QCheckBox(i18n("Bottom-Down View"), this);
-    bottomUpCheckbox->setToolTip(i18n("Enable the bottom-down flame graph view. When this is unchecked, "
+    m_backButton = new QPushButton(this);
+    m_backButton->setIcon(QIcon::fromTheme(QStringLiteral("go-previous")));
+    m_backButton->setToolTip(QStringLiteral("Go back in symbol view history"));
+    m_forwardButton = new QPushButton(this);
+    m_forwardButton->setIcon(QIcon::fromTheme(QStringLiteral("go-next")));
+    m_forwardButton->setToolTip(QStringLiteral("Go forward in symbol view history"));
+
+    auto bottomUpCheckbox = new QCheckBox(i18n("Bottom-Up View"), this);
+    bottomUpCheckbox->setToolTip(i18n("Enable the bottom-up flame graph view. When this is unchecked, "
                                       "the top-down view is enabled by default."));
+    bottomUpCheckbox->setChecked(m_showBottomUpData);
     connect(bottomUpCheckbox, &QCheckBox::toggled, this, [this, bottomUpCheckbox] {
         m_showBottomUpData = bottomUpCheckbox->isChecked();
         showData();
@@ -426,24 +501,66 @@ FlameGraph::FlameGraph(QWidget* parent, Qt::WindowFlags flags)
                 showData();
             });
 
-    m_displayLabel->setWordWrap(true);
-    m_displayLabel->setTextInteractionFlags(m_displayLabel->textInteractionFlags() | Qt::TextSelectableByMouse);
+    m_searchInput = new QLineEdit(this);
+    m_searchInput->setPlaceholderText(i18n("Search..."));
+    m_searchInput->setToolTip(i18n("<qt>Search the flame graph for a symbol.</qt>"));
+    m_searchInput->setClearButtonEnabled(true);
+    connect(m_searchInput, &QLineEdit::textChanged, this, &FlameGraph::setSearchValue);
 
     auto controls = new QWidget(this);
     controls->setLayout(new QHBoxLayout);
+    controls->layout()->addWidget(m_backButton);
+    controls->layout()->addWidget(m_forwardButton);
     controls->layout()->addWidget(m_costSource);
     controls->layout()->addWidget(bottomUpCheckbox);
     controls->layout()->addWidget(collapseRecursionCheckbox);
     controls->layout()->addWidget(costThreshold);
+    controls->layout()->addWidget(m_searchInput);
+
+    m_displayLabel->setWordWrap(true);
+    m_displayLabel->setTextInteractionFlags(m_displayLabel->textInteractionFlags() | Qt::TextSelectableByMouse);
+
+    m_searchResultsLabel->setWordWrap(true);
+    m_searchResultsLabel->setTextInteractionFlags(m_searchResultsLabel->textInteractionFlags()
+                                                  | Qt::TextSelectableByMouse);
+    m_searchResultsLabel->hide();
 
     setLayout(new QVBoxLayout);
     layout()->addWidget(controls);
     layout()->addWidget(m_view);
     layout()->addWidget(m_displayLabel);
+    layout()->addWidget(m_searchResultsLabel);
 
-    addAction(KStandardAction::back(this, SLOT(navigateBack()), this));
-    addAction(KStandardAction::forward(this, SLOT(navigateForward()), this));
-    setContextMenuPolicy(Qt::ActionsContextMenu);
+    m_backAction = KStandardAction::back(this, SLOT(navigateBack()), this);
+    addAction(m_backAction);
+    connect(m_backButton, &QPushButton::released, m_backAction, &QAction::trigger);
+
+    m_forwardAction = KStandardAction::forward(this, SLOT(navigateForward()), this);
+    addAction(m_forwardAction);
+    connect(m_forwardButton, &QPushButton::released, m_forwardAction, &QAction::trigger);
+
+    m_resetAction = new QAction(QIcon::fromTheme(QStringLiteral("go-first")), i18n("Reset View"), this);
+    m_resetAction->setShortcut(Qt::Key_Escape);
+    connect(m_resetAction, &QAction::triggered, this, [this]() { selectItem(0); });
+    addAction(m_resetAction);
+    updateNavigationActions();
+    m_view->setContextMenuPolicy(Qt::CustomContextMenu);
+    connect(m_view, &QWidget::customContextMenuRequested, this, [this](const QPoint& point) {
+        auto* menu = new QMenu(this);
+        menu->setAttribute(Qt::WA_DeleteOnClose, true);
+        if (auto item = static_cast<const FrameGraphicsItem*>(m_view->itemAt(point))) {
+            auto* action = menu->addAction(i18n("View Caller/Callee"));
+            connect(action, &QAction::triggered, this,
+                    [this, item]() { emit callerCalleeViewRequested(item->symbol()); });
+
+            auto* copy = menu->addAction(QIcon::fromTheme(QStringLiteral("edit-copy")), tr("Copy"));
+            connect(copy, &QAction::triggered, this, [item]() { qApp->clipboard()->setText(item->description()); });
+
+            menu->addSeparator();
+        }
+        menu->addActions(actions());
+        menu->popup(m_view->mapToGlobal(point));
+    });
 }
 
 FlameGraph::~FlameGraph() = default;
@@ -463,6 +580,7 @@ bool FlameGraph::eventFilter(QObject* object, QEvent* event)
                 }
                 m_selectedItem = m_selectionHistory.size();
                 m_selectionHistory.push_back(item);
+                updateNavigationActions();
             }
         }
     } else if (event->type() == QEvent::MouseMove) {
@@ -480,8 +598,22 @@ bool FlameGraph::eventFilter(QObject* object, QEvent* event)
             selectItem(m_selectionHistory.at(m_selectedItem));
         }
         updateTooltip();
-    } else if (event->type() == QEvent::Hide) {
-        setData(nullptr);
+    } else if (event->type() == QEvent::ToolTip) {
+        auto tooltip = m_displayLabel->toolTip();
+
+        if (m_tooltipItem != m_view->itemAt(m_view->mapFromGlobal(QCursor::pos()))) {
+            // don't show a tooltip when the cursor is in the empty region
+            tooltip.clear();
+        }
+
+        if (tooltip.isEmpty()) {
+            QToolTip::hideText();
+        } else {
+            QToolTip::showText(QCursor::pos(), QLatin1String("<qt>") + tooltip.toHtmlEscaped() + QLatin1String("</qt>"),
+                               this);
+        }
+        event->accept();
+        return true;
     }
     return ret;
 }
@@ -512,9 +644,12 @@ void FlameGraph::showData()
 {
     setData(nullptr);
 
-    m_buildingScene = true;
     using namespace ThreadWeaver;
     auto data = m_showBottomUpData ? m_bottomUpData : m_topDownData;
+    if (!data.resultData)
+        return;
+
+    m_buildingScene = true;
     bool collapseRecursion = m_collapseRecursion;
     auto source = m_costSource->currentData().value<CostType>();
     auto threshold = m_costThreshold;
@@ -541,8 +676,6 @@ void FlameGraph::updateTooltip()
     const auto text = m_tooltipItem ? m_tooltipItem->description() : QString();
     m_displayLabel->setToolTip(text);
     const auto metrics = m_displayLabel->fontMetrics();
-    // FIXME: the HTML text has tons of stuff that is not printed,
-    //        which lets the text get cut-off too soon...
     m_displayLabel->setText(metrics.elidedText(text, Qt::ElideRight, m_displayLabel->width()));
 }
 
@@ -550,10 +683,12 @@ void FlameGraph::setData(FrameGraphicsItem* rootItem)
 {
     m_scene->clear();
     m_buildingScene = false;
+    m_tooltipItem = nullptr;
     m_rootItem = rootItem;
     m_selectionHistory.clear();
     m_selectionHistory.push_back(rootItem);
     m_selectedItem = 0;
+    updateNavigationActions();
     if (!rootItem) {
         auto text = m_scene->addText(i18n("generating flame graph..."));
         m_view->centerOn(text);
@@ -567,11 +702,22 @@ void FlameGraph::setData(FrameGraphicsItem* rootItem)
     rootItem->setRect(0, 0, 800, m_view->fontMetrics().height() + 4);
     m_scene->addItem(rootItem);
 
+    if (!m_searchInput->text().isEmpty()) {
+        setSearchValue(m_searchInput->text());
+    }
+
     if (isVisible()) {
         selectItem(m_rootItem);
     }
 }
 
+void FlameGraph::selectItem(int item)
+{
+    m_selectedItem = item;
+    updateNavigationActions();
+    selectItem(m_selectionHistory.at(m_selectedItem));
+}
+
 void FlameGraph::selectItem(FrameGraphicsItem* item)
 {
     if (!item) {
@@ -604,18 +750,57 @@ void FlameGraph::selectItem(FrameGraphicsItem* item)
     setTooltipItem(item);
 }
 
+void FlameGraph::setSearchValue(const QString& value)
+{
+    if (!m_rootItem) {
+        return;
+    }
+
+    auto match = applySearch(m_rootItem, value);
+
+    if (value.isEmpty()) {
+        m_searchResultsLabel->hide();
+    } else {
+        QString label;
+        const auto costFraction = Util::formatCostRelative(match.directCost, m_rootItem->cost());
+        switch (m_costSource->currentData().value<CostType>()) {
+        case Allocations:
+        case Temporary:
+            label = i18n("%1 (%2% of total of %3) allocations matched by search.", match.directCost, costFraction,
+                         m_rootItem->cost());
+            break;
+        case Peak:
+        case Leaked:
+            label = i18n("%1 (%2% of total of %3) matched by search.", Util::formatBytes(match.directCost),
+                         costFraction, Util::formatBytes(m_rootItem->cost()));
+            break;
+        }
+        m_searchResultsLabel->setText(label);
+        m_searchResultsLabel->show();
+    }
+}
+
 void FlameGraph::navigateBack()
 {
     if (m_selectedItem > 0) {
-        --m_selectedItem;
+        selectItem(m_selectedItem - 1);
     }
-    selectItem(m_selectionHistory.at(m_selectedItem));
 }
 
 void FlameGraph::navigateForward()
 {
     if ((m_selectedItem + 1) < m_selectionHistory.size()) {
-        ++m_selectedItem;
+        selectItem(m_selectedItem + 1);
     }
-    selectItem(m_selectionHistory.at(m_selectedItem));
+}
+
+void FlameGraph::updateNavigationActions()
+{
+    const bool hasItems = m_selectedItem > 0;
+    const bool isNotLastItem = m_selectedItem + 1 < m_selectionHistory.size();
+    m_backAction->setEnabled(hasItems);
+    m_forwardAction->setEnabled(isNotLastItem);
+    m_resetAction->setEnabled(hasItems);
+    m_backButton->setEnabled(hasItems);
+    m_forwardButton->setEnabled(isNotLastItem);
 }
