@@ -12,6 +12,7 @@
 #include <KLocalizedString>
 
 #include "analyze/suppressions.h"
+#include "util/config.h"
 
 #include "gui_config.h"
 #include "mainwindow.h"
@@ -61,9 +62,9 @@ int main(int argc, char** argv)
 
     KLocalizedString::setApplicationDomain("heaptrack");
 
-    KAboutData aboutData(QStringLiteral("heaptrack_gui"), i18n("Heaptrack GUI"), QStringLiteral("0.1"),
-                         i18n("A visualizer for heaptrack data files."), KAboutLicense::LGPL,
-                         i18n("Copyright 2015, Milian Wolff <mail@milianw.de>"), QString(),
+    KAboutData aboutData(QStringLiteral("heaptrack_gui"), i18n("Heaptrack GUI"),
+                         QStringLiteral(HEAPTRACK_VERSION_STRING), i18n("A visualizer for heaptrack data files."),
+                         KAboutLicense::LGPL, i18n("Copyright 2015, Milian Wolff <mail@milianw.de>"), QString(),
                          QStringLiteral("mail@milianw.de"));
 
     aboutData.addAuthor(i18n("Milian Wolff"), i18n("Original author, maintainer"), QStringLiteral("mail@milianw.de"),
