@@ -1,20 +1,8 @@
 /*
- * Copyright 2016-2017 Milian Wolff <mail@milianw.de>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
- */
+    SPDX-FileCopyrightText: 2016-2017 Milian Wolff <mail@milianw.de>
+
+    SPDX-License-Identifier: LGPL-2.1-or-later
+*/
 
 #include "topproxy.h"
 
@@ -32,8 +20,6 @@ TreeModel::Columns toSource(TopProxy::Type type)
         return TreeModel::AllocationsColumn;
     case TopProxy::Temporary:
         return TreeModel::TemporaryColumn;
-    case TopProxy::Allocated:
-        return TreeModel::AllocatedColumn;
     }
     Q_UNREACHABLE();
 }
