@@ -17,6 +17,7 @@
 #include <unistd.h>
 
 #include <sys/mman.h>
+#include <limits.h>
 
 #include <type_traits>
 
@@ -26,10 +27,16 @@
  * @brief Experimental support for symbol overloading after runtime injection.
  */
 
+#if ULONG_MAX == 0xffffffffffffffff
+#define WORDSIZE 64
+#elif ULONG_MAX == 0xffffffff
+#define WORDSIZE 32
+#endif
+
 #ifndef ELF_R_SYM
-#if __WORDSIZE == 64
+#if WORDSIZE == 64
 #define ELF_R_SYM(i) ELF64_R_SYM(i)
-#elif __WORDSIZE == 32
+#elif WORDSIZE == 32
 #define ELF_R_SYM(i) ELF32_R_SYM(i)
 #else
 #error unsupported word size
@@ -37,15 +44,33 @@
 #endif
 
 #ifndef ElfW
-#if __WORDSIZE == 64
+#if WORDSIZE == 64
 #define ElfW(type) Elf64_##type
-#elif __WORDSIZE == 32
+#elif WORDSIZE == 32
 #define ElfW(type) Elf32_##type
 #else
 #error unsupported word size
 #endif
 #endif
 
+// NOTE: adding noexcept to C functions is a hard error in clang++
+//       (but not even a warning in GCC, even with -Wall)
+#if defined(__GNUC__) && !defined(__clang__)
+#define LIBC_FUN_ATTRS noexcept
+#else
+#define LIBC_FUN_ATTRS
+#endif
+
+extern "C" {
+
+// Foward declare mimalloc (https://github.com/microsoft/mimalloc) functions so we don't need to include its .h.
+__attribute__((weak)) void* mi_malloc(size_t size) LIBC_FUN_ATTRS;
+__attribute__((weak)) void* mi_calloc(size_t count, size_t size) LIBC_FUN_ATTRS;
+__attribute__((weak)) void* mi_realloc(void* p, size_t newsize) LIBC_FUN_ATTRS;
+__attribute__((weak)) void  mi_free(void* p) LIBC_FUN_ATTRS;
+
+}
+
 namespace {
 
 namespace Elf {
@@ -54,7 +79,7 @@ using Dyn = ElfW(Dyn);
 using Rel = ElfW(Rel);
 using Rela = ElfW(Rela);
 using Sym = ElfW(Sym);
-#if __WORDSIZE == 64
+#if WORDSIZE == 64
 using Sxword = ElfW(Sxword);
 using Xword = ElfW(Xword);
 #else
@@ -104,8 +129,10 @@ struct realloc
 
     static void* hook(void* ptr, size_t size) noexcept
     {
+        auto inPtr = reinterpret_cast<uintptr_t>(ptr);
         auto ret = original(ptr, size);
-        heaptrack_realloc(ptr, size, ret);
+        heaptrack_realloc2(inPtr, size, reinterpret_cast<uintptr_t>(ret));
+
         return ret;
     }
 };
@@ -183,6 +210,58 @@ struct posix_memalign
     }
 };
 
+// mimalloc functions
+struct mi_malloc
+{
+    static constexpr auto name = "mi_malloc";
+    static constexpr auto original = &::mi_malloc;
+
+    static void* hook(size_t size) noexcept
+    {
+        auto ptr = original(size);
+        heaptrack_malloc(ptr, size);
+        return ptr;
+    }
+};
+
+struct mi_free
+{
+    static constexpr auto name = "mi_free";
+    static constexpr auto original = &::mi_free;
+
+    static void hook(void* ptr) noexcept
+    {
+        heaptrack_free(ptr);
+        original(ptr);
+    }
+};
+
+struct mi_realloc
+{
+    static constexpr auto name = "mi_realloc";
+    static constexpr auto original = &::mi_realloc;
+
+    static void* hook(void* ptr, size_t size) noexcept
+    {
+        auto ret = original(ptr, size);
+        heaptrack_realloc(ptr, size, ret);
+        return ret;
+    }
+};
+
+struct mi_calloc
+{
+    static constexpr auto name = "mi_calloc";
+    static constexpr auto original = &::mi_calloc;
+
+    static void* hook(size_t num, size_t size) noexcept
+    {
+        auto ptr = original(num, size);
+        heaptrack_malloc(ptr, num * size);
+        return ptr;
+    }
+};
+
 template <typename Hook>
 bool hook(const char* symname, Elf::Addr addr, bool restore)
 {
@@ -220,7 +299,10 @@ void apply(const char* symname, Elf::Addr addr, bool restore)
         || hook<cfree>(symname, addr, restore)
 #endif
         || hook<posix_memalign>(symname, addr, restore) || hook<dlopen>(symname, addr, restore)
-        || hook<dlclose>(symname, addr, restore);
+        || hook<dlclose>(symname, addr, restore)
+        // mimalloc functions
+        || hook<mi_malloc>(symname, addr, restore) || hook<mi_free>(symname, addr, restore) || hook<mi_realloc>(symname, addr, restore)
+        || hook<mi_calloc>(symname, addr, restore);
 }
 }
 
@@ -228,13 +310,13 @@ template <typename T, Elf::Sxword AddrTag, Elf::Sxword SizeTag>
 struct elftable
 {
     using type = T;
-    T* table = nullptr;
-    Elf::Xword size = {};
+    Elf::Addr table = 0;
+    Elf::Xword size = 0;
 
     bool consume(const Elf::Dyn* dyn) noexcept
     {
         if (dyn->d_tag == AddrTag) {
-            table = reinterpret_cast<T*>(dyn->d_un.d_ptr);
+            table = dyn->d_un.d_ptr;
             return true;
         } else if (dyn->d_tag == SizeTag) {
             size = dyn->d_un.d_val;
@@ -242,19 +324,34 @@ struct elftable
         }
         return false;
     }
+
+    explicit operator bool() const noexcept
+    {
+        return table && size;
+    }
+
+    T* start(Elf::Addr tableOffset) const noexcept
+    {
+        return reinterpret_cast<T*>(table + tableOffset);
+    }
+
+    T* end(Elf::Addr tableOffset) const noexcept
+    {
+        return reinterpret_cast<T*>(table + tableOffset + size);
+    }
 };
 
 using elf_string_table = elftable<const char, DT_STRTAB, DT_STRSZ>;
 using elf_rel_table = elftable<Elf::Rel, DT_REL, DT_RELSZ>;
 using elf_rela_table = elftable<Elf::Rela, DT_RELA, DT_RELASZ>;
 using elf_jmprel_table = elftable<Elf::Rela, DT_JMPREL, DT_PLTRELSZ>;
-using elf_symbol_table = elftable<Elf::Sym, DT_SYMTAB, DT_SYMENT>;
+using elf_symbol_table = elftable<const Elf::Sym, DT_SYMTAB, DT_SYMENT>;
 
 template <typename Table>
 void try_overwrite_elftable(const Table& jumps, const elf_string_table& strings, const elf_symbol_table& symbols,
                             const Elf::Addr base, const bool restore) noexcept
 {
-    const auto elf_base =
+    Elf::Addr tableOffset =
 #ifdef __linux__
         0; // Already has memory addresses
 #elif defined(__FreeBSD__)
@@ -262,13 +359,25 @@ void try_overwrite_elftable(const Table& jumps, const elf_string_table& strings,
 #else
 #error port me
 #endif
-    const auto rela_start = reinterpret_cast<typename Table::type*>(reinterpret_cast<char*>(jumps.table) + elf_base);
-    const auto rela_end = reinterpret_cast<typename Table::type*>(reinterpret_cast<char*>(jumps.table) + jumps.size + elf_base);
-    const auto sym_start = reinterpret_cast<Elf::Sym*>(reinterpret_cast<char*>(symbols.table) + elf_base);
-    const auto str_start = reinterpret_cast<const char*>(strings.table) + elf_base;
+
+    const auto rela_start = jumps.start(tableOffset);
+    const auto rela_end = jumps.end(tableOffset);
+
+    const auto sym_start = symbols.start(tableOffset);
+    // we have no total size of the symbol table, so we cannot test that for validity
+
+    const auto str_start = strings.start(tableOffset);
+    const auto str_end = strings.end(tableOffset);
+    const auto num_str = static_cast<uintptr_t>(str_end - str_start);
+
     for (auto rela = rela_start; rela < rela_end; rela++) {
-        const auto index = ELF_R_SYM(rela->r_info);
-        const char* symname = str_start + sym_start[index].st_name;
+        const auto sym_index = ELF_R_SYM(rela->r_info);
+        const auto str_index = sym_start[sym_index].st_name;
+        if (str_index < 0 || str_index >= num_str)
+            continue;
+
+        const char* symname = str_start + str_index;
+
         auto addr = rela->r_offset + base;
         hooks::apply(symname, addr, restore);
     }
@@ -287,10 +396,22 @@ void try_overwrite_symbols(const Elf::Dyn* dyn, const Elf::Addr base, const bool
         symbols.consume(dyn) || strings.consume(dyn) || rels.consume(dyn) || relas.consume(dyn) || jmprels.consume(dyn);
     }
 
+    if (!symbols || !strings) {
+        return;
+    }
+
     // find symbols to overwrite
-    try_overwrite_elftable(rels, strings, symbols, base, restore);
-    try_overwrite_elftable(relas, strings, symbols, base, restore);
-    try_overwrite_elftable(jmprels, strings, symbols, base, restore);
+    if (rels) {
+        try_overwrite_elftable(rels, strings, symbols, base, restore);
+    }
+
+    if (relas) {
+        try_overwrite_elftable(relas, strings, symbols, base, restore);
+    }
+
+    if (jmprels) {
+        try_overwrite_elftable(jmprels, strings, symbols, base, restore);
+    }
 }
 
 int iterate_phdrs(dl_phdr_info* info, size_t /*size*/, void* data) noexcept
