@@ -24,11 +24,35 @@ using namespace std;
 #define HAVE_ALIGNED_ALLOC 0
 #endif
 
+// NOTE: adding noexcept to C functions is a hard error in clang++
+//       (but not even a warning in GCC, even with -Wall)
+#if defined(__GNUC__) && !defined(__clang__)
+#define LIBC_FUN_ATTRS noexcept
+#else
+#define LIBC_FUN_ATTRS
+#endif
+
+extern "C" {
+
+// Foward declare mimalloc (https://github.com/microsoft/mimalloc) functions so we don't need to include its .h.
+void* mi_malloc(size_t size) LIBC_FUN_ATTRS;
+void* mi_calloc(size_t count, size_t size) LIBC_FUN_ATTRS;
+void* mi_realloc(void* p, size_t newsize) LIBC_FUN_ATTRS;
+void  mi_free(void* p) LIBC_FUN_ATTRS;
+
+}
+
 namespace {
 
 namespace hooks {
 
-template <typename Signature, typename Base>
+enum class HookType
+{
+    Required,
+    Optional
+};
+
+template <typename Signature, typename Base, HookType Type>
 struct hook
 {
     Signature original = nullptr;
@@ -36,6 +60,9 @@ struct hook
     void init() noexcept
     {
         auto ret = dlsym(RTLD_NEXT, Base::identifier);
+        if (!ret && Type == HookType::Optional) {
+            return;
+        }
         if (!ret) {
             fprintf(stderr, "Could not find original function %s\n", Base::identifier);
             abort();
@@ -55,8 +82,8 @@ struct hook
     }
 };
 
-#define HOOK(name)                                                                                                     \
-    struct name##_t : public hook<decltype(&::name), name##_t>                                                         \
+#define HOOK(name, type)                                                                                               \
+    struct name##_t : public hook<decltype(&::name), name##_t, type>                                                   \
     {                                                                                                                  \
         static constexpr const char* identifier = #name;                                                               \
     } name
@@ -64,22 +91,28 @@ struct hook
 #pragma GCC diagnostic push
 #pragma GCC diagnostic ignored "-Wignored-attributes"
 
-HOOK(malloc);
-HOOK(free);
-HOOK(calloc);
+HOOK(malloc, HookType::Required);
+HOOK(free, HookType::Required);
+HOOK(calloc, HookType::Required);
 #if HAVE_CFREE
-HOOK(cfree);
+HOOK(cfree, HookType::Optional);
 #endif
-HOOK(realloc);
-HOOK(posix_memalign);
+HOOK(realloc, HookType::Required);
+HOOK(posix_memalign, HookType::Optional);
 #if HAVE_VALLOC
-HOOK(valloc);
+HOOK(valloc, HookType::Optional);
 #endif
 #if HAVE_ALIGNED_ALLOC
-HOOK(aligned_alloc);
+HOOK(aligned_alloc, HookType::Optional);
 #endif
-HOOK(dlopen);
-HOOK(dlclose);
+HOOK(dlopen, HookType::Required);
+HOOK(dlclose, HookType::Required);
+
+// mimalloc functions
+HOOK(mi_malloc, HookType::Optional);
+HOOK(mi_calloc, HookType::Optional);
+HOOK(mi_realloc, HookType::Optional);
+HOOK(mi_free, HookType::Optional);
 
 #pragma GCC diagnostic pop
 #undef HOOK
@@ -152,6 +185,12 @@ void init()
                        hooks::aligned_alloc.init();
 #endif
 
+                       // mimalloc functions
+                       hooks::mi_malloc.init();
+                       hooks::mi_calloc.init();
+                       hooks::mi_realloc.init();
+                       hooks::mi_free.init();
+
                        // cleanup environment to prevent tracing of child apps
                        unsetenv("LD_PRELOAD");
                        unsetenv("DUMP_HEAPTRACK_OUTPUT");
@@ -165,14 +204,6 @@ extern "C" {
 
 /// TODO: memalign, pvalloc, ...?
 
-// NOTE: adding noexcept to C functions is a hard error in clang++
-//       (but not even a warning in GCC, even with -Wall)
-#if defined(__GNUC__) && !defined(__clang__)
-#define LIBC_FUN_ATTRS noexcept
-#else
-#define LIBC_FUN_ATTRS
-#endif
-
 void* malloc(size_t size) LIBC_FUN_ATTRS
 {
     if (!hooks::malloc) {
@@ -340,4 +371,64 @@ int dlclose(void* handle) LIBC_FUN_ATTRS
 
     return ret;
 }
+
+// mimalloc functions, implementations just copied from above and names changed
+void* mi_malloc(size_t size) LIBC_FUN_ATTRS
+{
+    if (!hooks::mi_malloc) {
+        hooks::init();
+    }
+
+    void* ptr = hooks::mi_malloc(size);
+    heaptrack_malloc(ptr, size);
+    return ptr;
+}
+
+void* mi_realloc(void* ptr, size_t size) LIBC_FUN_ATTRS
+{
+    if (!hooks::mi_realloc) {
+        hooks::init();
+    }
+
+    void* ret = hooks::mi_realloc(ptr, size);
+
+    if (ret) {
+        heaptrack_realloc(ptr, size, ret);
+    }
+
+    return ret;
+}
+
+void* mi_calloc(size_t num, size_t size) LIBC_FUN_ATTRS
+{
+    if (!hooks::mi_calloc) {
+        hooks::init();
+    }
+
+    void* ret = hooks::mi_calloc(num, size);
+
+    if (ret) {
+        heaptrack_malloc(ret, num * size);
+    }
+
+    return ret;
+}
+
+void mi_free(void* ptr) LIBC_FUN_ATTRS
+{
+    if (!hooks::mi_free) {
+        hooks::init();
+    }
+
+    if (hooks::dummyPool().isDummyAllocation(ptr)) {
+        return;
+    }
+
+    // call handler before handing over the real free implementation
+    // to ensure the ptr is not reused in-between and thus the output
+    // stays consistent
+    heaptrack_free(ptr);
+
+    hooks::mi_free(ptr);
+}
 }
